#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include "bxlMsq.h"


int bxlMsqOpen (const int msqKey, const int queue_size)
{
	struct msqid_ds buf;
	int msqID;

	msqID = -1;

	if ((msqID = msgget(msqKey, IPC_CREAT | 0666)) < 0) {
		return EXIT_FAILURE;
	}

	if (queue_size > 0 && msgctl(msqID, IPC_STAT, &buf) == 0) {
		if (buf.msg_qbytes < queue_size) {
			buf.msg_qbytes = queue_size;
			msgctl(msqID, IPC_SET, &buf);
		}
	}

	return msqID;
}

int bxlMsqSend (
		const int writeMsqID, 
		const int mtype, 
		const char *data, 
		const int len)
{
	struct bxlMsg_t *msg;
	int msgLen;
	int ret;

	if (data == NULL || len <= 0) {
		printf("Incorrect parameter.\n");

		return EXIT_FAILURE;
	}

	msgLen = sizeof(long) + len;

	if ((msg = (struct bxlMsg_t *)calloc(msgLen, sizeof(char))) == NULL) {
		printf("Failed memory allocation.\n");

		return EXIT_FAILURE;
	}

	msg->mtype  = mtype;
//	strncpy(msg->mtext, data, len);
//	msg->mlen	= len;
	memcpy(msg->mtext, data, len);
	ret = msgsnd(writeMsqID, msg, msgLen, IPC_NOWAIT);

	
	/*
	{
		int i = 0;

		printf("balMsqSend() len = %d\n", len);
//		fprintf(stdout, "bxlMsqSend() msg->mtext = ");
		for (i = 0; i < len; i++) {
			printf("%c (0x%x)", *(msg->mtext + i), *(msg->mtext + i));
		}

		printf("\n");
	}
	*/
	

	free(msg);

	if (ret < 0) {
		printf("Failed to send msq.\n");

		return EXIT_FAILURE;
	} else {
		return EXIT_SUCCESS;
	}
}

struct bxlMsg_t *bxlMsqRecv (
		const int readMsqID, 
		const int mtype, 
		const int expectSize, 
		int *len)
{
	int nread;
	int mSize;
	struct bxlMsg_t *msg;

	nread = 0;
	mSize = expectSize;
	msg = NULL;

	if (readMsqID < 0 || expectSize <= 0) {
		*len = -1;

		return NULL;
	}

msq_recv_retry:
	msg = realloc(msg, sizeof(long) + mSize + 1);
	if (msg == NULL) {
		printf("fail memory allocation.\n");

		return NULL;
	}

	nread = msgrcv(readMsqID, msg, mSize, mtype, IPC_NOWAIT);

	if (nread < 0) {
		if (errno == ENOMSG || errno == EAGAIN) {
			*len = 0;
		} else if (errno == E2BIG) {
			mSize *= 2;

			goto msq_recv_retry;
		} else {
			*len = -1;
		}

		free(msg);

		return NULL;
	} else {
		*len = nread;
		msg->mtext[nread] = '\0';

		return msg;
	}
}

int bxlMsqClean (const int msqid)
{
	struct msqid_ds buf; 
	char    *ptr;
	int 	cnt;
	struct timespec req;

	if (msqid < 0) {
		printf("bxlMsqClean() : Incorrect parameter.\n");

		return EXIT_FAILURE;
	}

	if (msgctl(msqid, IPC_STAT, &buf) == 0) { 
		cnt = buf.msg_qnum;

		while (cnt > 0) {
			ptr = calloc(buf.msg_cbytes, sizeof(char));
			msgrcv(msqid, ptr, buf.msg_cbytes, 0, IPC_NOWAIT | MSG_NOERROR);
			cnt--;

			free(ptr);
		}

		req.tv_sec = 0;
		req.tv_nsec = 30000;
		nanosleep(&req, NULL);

		return EXIT_SUCCESS;
	} else {
		return EXIT_FAILURE;  
	}    
}

int bxlRMMsq (const int msqid)
{
	if (msqid < 0) {
		return EXIT_FAILURE;
	}

	return msgctl(msqid, IPC_RMID, NULL);
}
